////////////////////////////////////////////////////////////////////// 
// libsrc/MathNormal.cpp 
// (c) 2000-2001 Goncalo Abecasis
// 
// This file is distributed as part of the MERLIN source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile MERLIN.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Monday August 05, 2002
// 
 
#include "MathNormal.h"
#include "MathSVD.h"
#include "MathGenMin.h"
#include "MathStats.h"

#include <math.h>

NormalEquations::NormalEquations() :
   variances("variances"), means("linear"),
   linearModel("linear"), scores("scores"),
   cholesky(), varMatrix("variances"), residuals("residuals"),
   Qi("Qi")
   {
   variances.Dimension(0);
   varComponents = NULL;
   includeLikelihoodConstant = true;
   preliminary = false;
   multiple = 1;
   }

NormalEquations::~NormalEquations()
   {
   Free();
   }

void NormalEquations::Free()
   {
   if (variances.dim)
      delete [] varComponents;
   varComponents = NULL;
   }

void NormalEquations::CalculateResiduals()
   {
   // deviation from predicted scores in linear model
   residuals.Product(linearModel, means);
   residuals.Subtract(scores);
   }

void NormalEquations::CalculateCovariances()
   {
   // Construct the covariance matrix by summing its
   // component parts and then multiply the residuals
   // by inverse of covariance matrix
   //    - Inv(A)*b = x where A*x = b and x = BackSubst(b)
   //    - cholesky decomposition only requires the upper
   //      triangle of the positive definite A
   varMatrix.Zero();
   for (int j = 0; j < variances.dim; j++)
      for (int r = 0; r < varMatrix.rows; r++)
         for (int c = r; c < varMatrix.rows; c++)
            varMatrix[r][c] += variances[j] * varComponents[j][r][c];

   /* The following two lines achieve the same in a
      prettier but more inefficient manner:

      for (int j = 0; j < variances.dim; j++)
         varMatrix.AddMultiple(variances[j], varComponents[j]);
   */
   }

double NormalEquations::Evaluate()
   {
   // (2pi)^(-k/2)
   likelihood = includeLikelihoodConstant ? constant : 0.0;

   // deviation from predicted scores in linear model
   CalculateResiduals();

   // The variance matrix A (varMatrix)
   CalculateCovariances();

   // and its cholesky decomposition are updated
   cholesky.FastDecompose(varMatrix);

   // This is cholesky.x = Inv(A) * residuals
   cholesky.BackSubst(residuals);

   // Add e ^ [- 0.5 * (yi - ui)T * A^-1 * (yi - ui)]
   likelihood -= 0.5 * residuals.InnerProduct(cholesky.x);

   // Det(A[i]^-1/2) = -Det(sqrt(A)) = -Det(L)
   likelihood -= cholesky.lnDeterminantL();

   return likelihood;
   }

void NormalEquations::Diagnostics()
   {
   CalculateResiduals();
   CalculateCovariances();

   cholesky.BackSubst(residuals);

   rawQ = residuals.InnerProduct(cholesky.x);
   Q = sqrt(2*rawQ) - sqrt(2 * residuals.dim - 1);

   Qi.Dimension(residuals.dim);

   Matrix M;
   Vector v, r;
   Cholesky chol;

   for (int i = 0; i < residuals.dim; i++)
      {
      M = varMatrix;

      v = M[i];
      M.DeleteColumn(i);
      M.DeleteRow(i);
      v.DeleteDimension(i);
      chol.Decompose(M);

      chol.BackSubst(v);

      double var = varMatrix[i][i] - v.InnerProduct(chol.x);

      r = residuals;
      r.DeleteDimension(i);

      chol.BackSubst(r);
      Qi[i] = residuals[i] - v.InnerProduct(chol.x);
      Qi[i] = Qi[i] * Qi[i] / var;
      }
   }

void NormalEquations::Prepare()
   {
   varMatrix.Dimension(scores.dim, scores.dim);
   residuals.Dimension(scores.dim);
   means.Dimension(linearModel.cols);
   constant = log(2 * M_PI) * -0.5 * scores.dim;
   }

void NormalEquations::SelectPoint(Vector & v)
   {
   for (int i = 0; i < means.dim; i++)
      means[i] = v[i];

   for (int i = 0; i < variances.dim; i++)
      if (v[i + means.dim] > 16 || v[i] < -16)
         variances[i] = v[i + means.dim] > 0 ? 1e7 : 1e-7;
      else
         variances[i] = exp(v[i + means.dim]);
   }

void NormalEquations::RetrievePoint(Vector & v)
   {
   for (int i = 0; i < means.dim; i++)
      v[i] = means[i];
   for (int i = 0; i < variances.dim; i++)
      v[i + means.dim] = log(variances[i]);
   }

void NormalEquations::Dimension(int vcCount)
   {
   if (variances.dim != vcCount)
      {
      Free();

      varComponents = new Matrix[vcCount];
      variances.Dimension(vcCount);
      }
   }

bool NormalEquations::operator == (const NormalEquations & rhs)
   {
   if (scores != rhs.scores) return false;
   if (linearModel != rhs.linearModel) return false;
   for (int i = 0; i < variances.dim; i++)
      if (varComponents[i] != rhs.varComponents[i])
         return false;
   return true;
   }

// Normal Set class
//

NormalSet::NormalSet(int threads)
   : variances("variances"), means("means")
   {
   maxThreads = threads;
   count = size = 0;
   numericMinimizer = 0;
   precision = 1e-8;
   preliminary = iterative = false;
   }

void NormalSet::Free()
   {
   if (size)
      {
      for (int i = 0; i < size; i++)
         delete sets[i];
      delete [] sets;
      }
   size = 0;
   }

void NormalSet::Dimension(int setCount, int vcCount)
   {
   if (setCount > size)
      {
      Free();

      size = setCount;
      sets = new NormalEquations * [size];

      AllocateSets();
      }

   count = setCount;
   for (int i = 0; i < count; i++)
      sets[i]->Dimension(vcCount);
   }

void NormalSet::AllocateSets()
   {
   for (int i = 0; i < size; i++)
      sets[i] = new NormalEquations;
   }

double NormalSet::Evaluate()
   {
   for (int i = 0; i < count; i++)
        sets[i]->Evaluate();

   likelihood = 0;
   for (int i = 0; i < count; i++)
      likelihood -= sets[i]->likelihood;

   return likelihood;
   }

void * NormalSet::EvaluateOneSet(void * which)
   {
   NormalEquations * w = (NormalEquations *) which;

   w->Evaluate();

   return NULL;
   }

void NormalSet::SelectPoint(Vector & v)
   {
   for (int i = 0; i < count; i++)
      sets[i]->SelectPoint(v);

   variances = sets[0]->variances;
   means = sets[0]->means;
   }

void NormalSet::Solve()
   {
   EditLinearDegenerates();

   GeneralMinimizer * solver;

   switch (numericMinimizer)
      {
      case NORMAL_AMOEBA_MIN : solver = new AmoebaMinimizer(); break;
      case NORMAL_POWELL_MIN : solver = new PowellMinimizer(); break;
      }

   solver->func = new NormalSolver(this);

   // The iterative flag denotes really tough problems, like
   // discrete data. Use all the tricks in the book to try
   // and save time...

   // if (iterative) RemoveRedundancy();

   for (int round = !iterative; round <= 1; round++)
      {
      // For solving discrete data problems, a preliminary
      // round estimates the mean parameters only
      preliminary = round == 0;

      for (int i = 0; i < count; i++)
         sets[i]->preliminary = preliminary;

      // set the number of parameters to minimize
      int parameters = CountParameters();
      solver->Reset(parameters);

      // Select starting point
      GetStartingPoint(solver->point);

      // Two rounds of minimizing to be safe
      solver->Minimize(precision);
      if (!iterative) solver->Reset(parameters);
      if (!iterative) solver->Minimize(precision);

      SelectPoint(solver->point);
      }

   delete solver->func;
   delete solver;
   }

int NormalSet::CountObservations()
   {
   int rows = 0;
   for (int i = 0; i < count; i++)
      rows += sets[i]->linearModel.rows;

   return rows;
   }

void NormalSet::EditLinearDegenerates()
   {
   // create a matrix M with the linear side of the model
   // for each of the normal equations
   int    rows = CountObservations();
   Matrix m(rows, sets[0]->linearModel.cols);;
   Vector b(rows);

   for (int i = 0, total = 0; i < count; i++)
      for (int j = 0; j < sets[i]->linearModel.rows; j++)
         {
         b[total] = sets[i]->scores[j];
         m[total++] = sets[i]->linearModel[j];
         }

   // now we decompose m successively and remove redundant
   // columns from m and each individual set...
   SVD engine;

   for (int col = 1; col <= m.cols; )
      {
      engine.Decompose(m, m.rows, col);
      engine.Edit();

      bool redundant = false;
      for (int j = 0; j < col; j++)
         if (engine.w[j] == 0.0)
            redundant = true;

      if (redundant)
         {
         for (int i = 0; i < count; i++)
            sets[i]->linearModel.DeleteColumn(col - 1);
         m.DeleteColumn(col - 1);

         // Check if last column was deleted ...
         if (col > m.cols) engine.Decompose(m, m.rows, m.cols);
         }
      else
         col++;
      }

  // at this point we might as well use the SVD information
  // to choose good starting points
  for (int i = 0; i < count; i++)
     sets[i]->Prepare();

  variances.Dimension(sets[0]->variances.dim);

  engine.BackSubst(b);

  // Assume all variances * are * the total variance
  // and that SVD estimates are good for the linear side

  means = engine.x;
  means.Dimension(m.cols);

  // make sure the variances are positive despite roundoff
  double _TINY_ = 1e-20;
  double residualVar = fabs(engine.RSS(m, b)) + _TINY_;
  residualVar /= rows;

  variances.Set(residualVar / variances.dim);
  }

int NormalSet::CountParameters()
   {
   return means.dim + variances.dim;
   };

void NormalSet::GetStartingPoint(Vector & startPoint)
   {
   sets[0]->variances = variances;
   sets[0]->means = means;

   sets[0]->RetrievePoint(startPoint);
   }

void NormalSet::RemoveRedundancy()
   {
   for (int i = count - 1; i >= 0; i--)
      for (int j = 0; j < i; j++)
         if (*(sets[i]) == *(sets[j]))
            {
            sets[j]->multiple++;
            count--;
            NormalEquations * temp = sets[i];
            for ( j = i; j < count; j++)
               sets[j] = sets[j + 1];
            sets[j] = temp;
            }
   }

// Normal Solver classes
//

double NormalSolver::Evaluate(Vector & point)
   {
   normal->SelectPoint(point);
   return normal->Evaluate();
   };


 
